﻿package ru.k0t0vich.fluent 
{
	import flash.utils.getQualifiedClassName;
	import ru.k0t0vich.fluent.conditions.Condition;
	import ru.k0t0vich.fluent.parsers.FluentXMLParser;
   
	/**
	 * Класс реализующий структуру данных: "Естественный список." (с)k0t0vich (smile:))
	 * Представляет собой список элементов с методами выборки и сортировки организованными на "естественном языке".
	 * Позволяет: выбирать элементы удовлетворяющиее условиям, производить групповые операции над элементами выборки,
	 * сортировать выборки по полям элементов, добавлять/удалять элементы из списка.
	 * TODO: объединение/ разность списков.
	 * Выражение для выборки должны быть представлены в ДНФ(дизъюнктивно нормальная форма), либо сгруппированы.
	 * @exampleText
	 * 1)Для всех юнитов с Z>-80 и Z<-50 c уровнем не меньше 5 или для юнитов с уровнем 1 - отсортировать по Z и стрелять пушками типа 1 частотой 10.
	 * @example
	 * 
		//Создаем список юнитов
			var enemies:FluentList = new FluentList();
			
			// Заполняем
			for (var i:int = 0; i < 10; i++) 
			{
				var unit:Unit = new Unit(i,-i*10);
				enemies.push(unit);
			}
			// весь список
			trace("TOTAL: " + enemies);
			// выполняем
			enemies
			.where(Unit.Z, Condition.GT, -80)
			.where(Unit.Z, Condition.LT, -50)
			.where(Unit.LEVEL, Condition.GE,5)
				.or()
			.where("level", "==", 1)
				.sortOn("z",Array.NUMERIC)
				.call(Unit.DO_SHOOT,1,10);
	 * 
	 * 
	 * 
	 * 
	 * 
	 * 
	 * 
	 * 
	 * 
	 * 
	 * 
	 * 
	 * 
	 * @author k0t0vich
	 */
	public class FluentList
	{
		private var unitList:Array;
		
		private var dnfArray:Array = [];	
		private function get currConjunctionArray():Array { return dnfArray[dnfArray.length - 1] as Array; }
		
		public function FluentList($unitList:Array=null) 
		{
			if ($unitList) unitList = $unitList;
			else unitList = [];
			// начинаем дизъюнкцию
			clearConditions();
		}
		
		
		/**
		 * Синоним join()
		 * Добавить блок конъюнктов. 
		 * Дизъюнктивно объеденить с предыдущим блоком условий  (опрация ИЛИ)
		 * @return
		 */
		public function or():FluentList
		{
			dnfArray.push([]);
			return this;
		}
		

		/**
		 * Добавить в текущий конъюнкт условие (операция И)
		 * Валидность условия проверяется в рантайме.
		 * @param	$field  			поле элемента списка
		 * @param	$condition			условие (==,!=,<,>,<=,>=)	
		 * @param	$value				значение
		 * @return
		 */
		public function where($field:String, $condition:String, $value:*):FluentList
		{
			//trace("FluentList.where > $field : " + $field + ", $condition : " + $condition + ", $value : " + $value);
			currConjunctionArray.push(new Condition($field, $condition, $value));
			return this;
		}
		
		/**
		 * Синоним or() 
		 * Добавить блок конъюнктов. 
		 * Дизъюнктивно объеденить с предыдущим блоком условий  (опрация ИЛИ)
		 * @return
		 */
		public function join():FluentList
		{
			return or();
		}

		/**
		 *  Выбрать элементы удовлетворяюшие ДНФ
		 * @return
		 */		
		public function group():FluentList
		{
			var retFluentList:FluentList = new FluentList();
			var len:int = unitList.length;
			for (var i:int = 0; i < len; i++) 
			{
				var unit:* = unitList[i];
				if (valueConditionsWithUnit(unit))
				{
					retFluentList.push(unit);
				}
				
			}
			clearConditions();
			return retFluentList;
		}
		
		/**
		 * Очистить ДНФ
		 * @return
		 */
		private function clearConditions():FluentList
		{
			dnfArray = [];
			return or();
		}
		
		/**
		 * Разбираем ДНФ для юнита
		 * @param	unit
		 * @return
		 */
		private function valueConditionsWithUnit(unit:*):Boolean
		{
			var dnfLen:int = dnfArray.length;
			var dnfRet:Boolean = false;
			for (var i:int = 0; i < dnfLen; i++) 
			{
				var conjArray:Array = dnfArray[i] as Array;
				var conjLen:int = conjArray.length;
				var conjRet:Boolean = true;
				
				for (var j:int = 0; j < conjLen; j++) 
				{
					var condition:Condition =  conjArray[j] as Condition;
					if (!condition.valueCondition(unit))
					{
						conjRet = false;
						//досрочный выход из конъюнкта
						break;
					}
				}
				dnfRet = dnfRet || conjRet;
				if (dnfRet)	return true; // досрочный выход из дизъюнкта
			}
			return false;
		}
		
		/**
		 * Выполнить у юнитов удовлетворяющих ДНФ ф-цию
		 * @param	func : имя ф-ции
		 * @param	args : аргументы
		 */
		public function call(funcName:String,args:Array):FluentList
		{
			var len:int = unitList.length;
			for (var i:int = 0; i < len; i++) 
			{
				var unit:* = unitList[i];
				if (valueConditionsWithUnit(unit))
				{
					// действие над каждым отфильтрованным элементом
					try {
						var func:Function = unit[funcName];
					}
					catch (e:Error)
					{
						throw new Error("'"+funcName+"'"+ " is not member of "+ getQualifiedClassName(unit)+"! Standart error: "+e);
					}
					func.apply(unit, args);
				}
				
			}
			return clearConditions();
		}
		
		/**
		 * Выполнить ф-цию  над юнитами удовлетворяющими ДНФ
		 * @param	func : ф-ция вида func(unit:*,args)
		 * @param	args : аргументы
		 */
		public function execute(func:Function,...args):FluentList
		{
			var len:int = unitList.length;
			for (var i:int = 0; i < len; i++) 
			{
				var unit:* = unitList[i];
				if (valueConditionsWithUnit(unit))
				{
					func(unit, args);
				}
				
			}
			return clearConditions();
		}
		
		
		/**
		 * Распарсить XML сценария
		 * @param	xml
		 * @return
		 */
		public function parseXML(xml:XML):FluentList
		{
			return FluentXMLParser.parse(this,xml);
		}
		
		/**
		 * Композиционные методы для массива
		 * TODO: добавить остальные
		 */
		public function push(unit:*):FluentList
		{
			unitList.push(unit);
			return this;
		}
		
		
		public function pop():*
		{
			
			return unitList.pop();
		}
		
		public function get length():int
		{
			return unitList.length;
		}
		
		
		public function toString():String
		{
			return "[FluentList" + String(unitList) + " ]";
			
		}
		
		public function sortOn(names:*,options:*=0):FluentList
		{
			unitList.sortOn(names, options);
			return group();
		}
	}

}